program TAIL;

{$APPTYPE CONSOLE}

uses Windows, SysUtils;

const
  AppVer = 'TAIL.EXE v1.00 by WoodyBells.';

var
  FFileName:   string;
  FTail_F:     Boolean;
  FFileHandle: THandle;
  FShowRange:  Integer = 10;
  FCurrentPos: Int64 = 0;
  FFileSize:   Int64;
  FStdErr:     TextFile;

function Min(V1, V2: Int64): Int64;
begin
  if V1 < V2 then
    Result := V1
  else
    Result := V2; 
end;

function Max(V1, V2: Int64): Int64;
begin
  if V1 < V2 then
    Result := V2
  else
    Result := V1; 
end;

procedure AssignStdErr;
begin
  AssignFile(FStdErr, 'CONOUT$');
  ReWrite(FStdErr);
  TTextRec(FStdErr).Handle := Integer(GetStdHandle(STD_ERROR_HANDLE));
end;

procedure FreeStdErr;
begin
  CloseFile(FStdErr);
end;

procedure OutMsg(msg: string);
begin
  Write(msg);
end;

procedure OutErrMsg(msg: string);
begin
  WriteLn(FStdErr, msg);
end;

function GetLastErrorMessage: string;
begin
  Result := SysErrorMessage(GetLastError);
end;

function JTOpenFile(fName: PChar; DesiredAccess, ShareMode,
                    CreationDistribution, FlagsAndAttributes: DWord): THandle;
begin
  Result := CreateFile(fName, DesiredAccess, ShareMode, nil,
                       CreationDistribution, FlagsAndAttributes, 0);
end;

function JTCloseFile(f: THandle): Boolean;
begin
  Result := CloseHandle(f);
end;

function JTOpenFileRead(fName: PChar ): THandle;
begin
  Result := JTOpenFile(fName, GENERIC_READ,
                       FILE_SHARE_READ + FILE_SHARE_WRITE, OPEN_EXISTING,
                       FILE_ATTRIBUTE_ARCHIVE + FILE_ATTRIBUTE_COMPRESSED +
                       FILE_ATTRIBUTE_HIDDEN  + FILE_ATTRIBUTE_READONLY   +
                       FILE_ATTRIBUTE_SYSTEM  + FILE_ATTRIBUTE_TEMPORARY);
end;

function JTGetFileSize(f: THandle): Int64;
var
  dwSizeHigh, dwSizeLow: DWord;
begin
  dwSizeLow := GetFileSize(f, @dwSizeHigh);
  if (dwSizeLow = $FFFFFFFF) and (GetLastError <> NO_ERROR) then
    Result := 0
  else
    Result := dwSizeHigh * MAXDWORD + dwSizeLow;
end;

function JTGetFilePos(f: THandle): Int64;
begin
  Result := SetFilePointer(f, 0, nil, FILE_CURRENT);
end;

function JTFileSeek(f: THandle; Offset: Int64; MoveMethod: Word): Boolean;
var
   d: Int64;
begin
   d := SetFilePointer(f, Offset, nil, MoveMethod);
   Result := not ((d = $FFFFFFFF) and (GetLastError <> NO_ERROR));
end;

function ReadBytes: string;
var
  needsize, needsize2: Int64;
  rs: DWORD;
  pca: PChar;
begin
  Result := '';
  FFileSize := JTGetFileSize(FFileHandle);
  if FCurrentPos < 0 then FCurrentPos := 0;
  if FFileSize > FCurrentPos then
  begin
    needsize  := FFileSize - FCurrentPos;
    needsize2 := needsize;
    rs := Min(needsize, MAXDWORD);

    GetMem(pca, needsize + 1);
    try
      ZeroMemory(pca, needsize + 1);
      while needsize2 > 0 do
      begin
        JTFileSeek(FFileHandle, FCurrentPos, 0);
        FileRead(FFileHandle, pca^, rs);
        Result := Result + StrPas(pca);
        Dec(needsize2, MAXDWORD);
        rs := Min(needsize2, MAXDWORD);
      end;
    finally
      FreeMem(pca);
    end;
  end;
  FCurrentPos := FFileSize;
end;

procedure SleepTimer(ms: Byte);
var
  tick: Cardinal;
begin
  tick := GetTickCount;
  while GetTickCount - tick < ms do
  begin
    Sleep(10);
  end;
end;

function GetStartPosition: Int64;
var
  i, cnt, bufsize: Integer;
  spos, epos: Int64;
  buf: PChar;
begin
  spos := FFileSize - 50000;
  epos := FFileSize;
  cnt  := FShowRange;
  GetMem(buf, 1);
  try
    while cnt > 0 do
    begin
      Dec(spos, 50000);
      spos := Max(spos, 0);
      if epos < spos then Break;
      bufsize := epos - spos;
      ReallocMem(buf, bufsize);
      ZeroMemory(buf, bufsize);
      JTFileSeek(FFileHandle, spos, 0);
      FileRead(FFileHandle, buf^, bufsize);
      for i := bufsize - 1 downto 0 do
      begin
        if Byte(buf[i]) = $0A then // LF
        begin
          Dec(cnt);
          if cnt = 0 then
          begin
            spos := spos + i + 1;  
            Break;
          end;
        end;
      end;
      if cnt = 0 then Break;
      epos := spos - 1;
    end;
    Result := spos;
  finally
    FreeMem(buf);
  end;
end;

function ParamCheck: Boolean;
var
  i, wint: Integer;
  prm: string;
  showver: Boolean;
begin
  Result    := False;
  FTail_F   := False;
  FFileName := '';
  showver   := False;
  if ParamCount = 0 then
  begin
    OutErrMsg('USAGE: TAIL.EXE [-F] [-n] file_name');
    Exit;
  end;
  for i := 1 to ParamCount do
  begin
    prm := UpperCase(ParamStr(i));
    if prm[1] <> '-' then
      FFileName := ParamStr(i)
    else if prm = '-F' then
      FTail_F := True
    else if prm = '-V' then
    begin
      OutMsg(AppVer + #13#10);
      showver := True;
    end else
    begin
      wint := StrToIntDef(Copy(prm, 2, Length(prm) - 1), -1);
      if wint = -1 then
      begin
        OutErrMsg('MISSING OPTION(s)');
        Exit;
      end else
        FShowRange := wint;
    end;
  end;
  if (FFileName = '') or not FileExists(FFileName) then
  begin
    if not showver then
      OutErrMsg('MISSING FILENAME');
    Exit;
  end;
  Result := True;
end;


begin
  AssignStdErr;
  try
    if not ParamCheck then Exit;
    FFileHandle := JTOpenFileRead(PChar(FFileName));
    if FFileHandle = INVALID_HANDLE_VALUE then
    begin
      OutErrMsg('FILE OPEN ERROR!');
      Exit;
    end;
    try
      FFileSize := JTGetFileSize(FFileHandle);
      FCurrentPos := GetStartPosition;
      JTFileSeek(FFileHandle, FCurrentPos, 0);
      OutMsg(ReadBytes);
      while FTail_F do
      begin
        SleepTimer(200);
        OutMsg(ReadBytes);
      end;
    finally
      JTCloseFile(FFileHandle);
    end;
  finally
    FreeStdErr;
  end;
end.
